// QBindManager.cpp: implementation of the QBindManager class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "Query32.h"
#include "QBindManager.h"
#include "QLogFile.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

QBindManager::QBindManager()
{
  conn_acc_id = 0;
	m_nIdleTimeout = 30;
}

QBindManager::~QBindManager()
{
  WIN32_GUFF();

  POSITION pos = conn_map.GetHeadPosition();
  QConnection *conn;

  while (pos != NULL)
    {
      conn = conn_map.GetNext( pos );
      conn->Unbind();
			delete conn;
    }
}

//////////////////////////////////////////////////////////////////////
// QBindManager::GetConnection
// Return connection details (host name, port, user DN and password)
// for the identified connection.
BOOL QBindManager::GetConnection
	( ConnID conn_id, CString &host, int &port, CString &user, CString &password )
{
	WIN32_GUFF();

  POSITION pos = conn_map.GetHeadPosition();
  QConnection *conn;

  while ( pos != NULL )
    {
      conn = conn_map.GetNext( pos );
      if ( conn->ID() == conn_id )
        // The requested connection already exists
        {
					conn->GetConnection( host, port, user, password );
					return TRUE;
        }
    }

	return FALSE;
}

//////////////////////////////////////////////////////////////////////
// Specify a directory connection. If the specified connection
// already exists, increment the reference count.
ConnID QBindManager::SpecifyConnection
  ( CString &host, int port, CString &user, CString &password )
{
  WIN32_GUFF();

  POSITION pos = conn_map.GetHeadPosition();
  QConnection *conn;

  while ( pos != NULL )
    {
      conn = conn_map.GetNext( pos );
      if ( conn->Compare( host, port, user, password ) == TRUE )
        // The requested connection already exists
        {
					conn->AddRef();
					return conn->ID();
        }
    }

  //
  // Need to create a new connection record
	ConnID conn_id = conn_acc_id;
  if ( conn_acc_id == LONG_MAX )
		 conn_acc_id = 0;
	else
		conn_acc_id++;

  conn = new QConnection( conn_id, host, port, user, password );
  conn_map.SetAt( conn_id, conn );
	return conn_id;
}

//////////////////////////////////////////////////////////////////////
// QBindManager::AddRef
// Increment the reference count for the identified connection.
void QBindManager::AddRef( ConnID conn_id )
{
	WIN32_GUFF();

	QConnection *conn;
	if ( conn_map.Lookup( conn_id, conn ) == FALSE )
		{
			//
			// Can't add a use reference on an unknown connection!
#ifdef _DEBUG
			TRACE
				( "QBindManager::AddRef called on an unknown connection identifier!\n" );
#endif 
			return;
		}

	conn->AddRef();
}

//////////////////////////////////////////////////////////////////////
// QBindManager::LoseRef
// Decrement the reference count for the identified connection.
// If the reference count has reached zero unbind the connection.
void QBindManager::LoseRef( ConnID conn_id )
{
	WIN32_GUFF();

	if ( conn_id == UNDEF_CONN_ID )
		return;

	QConnection *conn;
	if ( conn_map.Lookup( conn_id, conn ) == FALSE )
		{
			//
			// Can't take use reference away from an unknown connection!
#ifdef _DEBUG
			TRACE
				( "QBindManager::LoseRef\nCalled on an unknown connection identifier!\n" );
#endif 
			return;
		}

	if ( conn->LoseRef() == 0 )
		conn->Unbind();
}

//////////////////////////////////////////////////////////////////////
// QBindManager::BindAsync
// Initiate an asynchronous bind on a specified directory connection. 
BOOL QBindManager::BindAsync( ConnID id, QError *&err )
{
  WIN32_GUFF();

	if ( id == UNDEF_CONN_ID )
		return FALSE;

  POSITION pos = conn_map.GetHeadPosition();
  QConnection *conn;
  BOOL conn_exists = FALSE;

  while ( pos != NULL )
    {
      conn = conn_map.GetNext( pos );
      if ( conn->ID() == id )
        // The requested connection already exists
        {
          conn_exists = TRUE;
          break;
        }
    }

  if ( conn_exists )
    {
      if ( conn->Bound() )
				return TRUE;
    }

  //
  // Request an asynchronous bind. 
  return conn->BindAsync( err );
}

//////////////////////////////////////////////////////////////////////
// QBindManager::RebindAsync
// Rebind the connection.
BOOL QBindManager::RebindAsync( ConnID conn_id, QError *&err )
{
  WIN32_GUFF();

  POSITION pos = conn_map.GetHeadPosition();
  QConnection *conn;
  BOOL conn_exists = FALSE;

  while ( pos != NULL )
    {
      conn = conn_map.GetNext( pos );
      if ( conn->ID() == conn_id )
        // The requested connection id is OK
        {
          conn_exists = TRUE;
          break;
        }
    }

  if ( conn_exists == FALSE )
		{
#ifdef _DEBUG
			TRACE
				( "QBindManager::BindAsync called on a non-existent connection id!\n" );
#endif
			err = new QError( QERR_INTERNAL );
			return FALSE;
		}

	if ( conn->Bound() == TRUE )
		return TRUE;

	return conn->RebindAsync( err );
}

//////////////////////////////////////////////////////////////////////
// QBindManager::BindWait
// Wait for the identified asynchronous bind operation to complete.
BOOL QBindManager::BindWait( ConnID id, QError *&err )
{
  WIN32_GUFF();
  QConnection *conn;
  
  if ( conn_map.Lookup( id, conn ) == FALSE )
		{
			err = new QError( QERR_INTERNAL );
			return FALSE;
		}

  return conn->BindWait( err );
}

//////////////////////////////////////////////////////////////////////
// QBindManager::Bound
// Check the status of a directory connection.
BOOL QBindManager::Bound( ConnID id )
{
  WIN32_GUFF();
  QConnection *conn;
  
  if ( conn_map.Lookup( id, conn ) == FALSE )
    return FALSE;

  return conn->Bound();
}

//////////////////////////////////////////////////////////////////////
// QBindManager::Exists
// Check if a directory connection with the given details exists.
BOOL QBindManager::Exists
  ( ConnID &id, const CString &host, int port, const CString &user, const CString &passwd )
{
  WIN32_GUFF();
  QConnection *conn;
  
  if ( conn_map.Lookup( id, conn ) == FALSE )
    return FALSE;

  return conn->Bound();
}

//////////////////////////////////////////////////////////////////////
// QBindManager::GetLDAPContext
// Get LDAP context. Result shouldn't be stored as the pointer will change
// as the connection is closed (due to inactivity) and then re-opened
LDAP *QBindManager::GetLDAPContext( ConnID id )
{
  WIN32_GUFF();
  QConnection *conn;
  
  if ( conn_map.Lookup( id, conn ) == FALSE )
    return NULL;

  return conn->GetLDAPContext();
}

//////////////////////////////////////////////////////////////////////
// QBindManager::Dump
// Dump a connection's details to the log file
void QBindManager::Dump( ConnID id )
{
	WIN32_GUFF();

	//
	// Find the connection record and dump it to the log file
	POSITION pos = conn_map.GetHeadPosition();
  QConnection *conn;
  while ( pos != NULL )
    {
      conn = conn_map.GetNext( pos );
      if ( conn->ID() == id )
				{
					conn->Dump();
					return;
				}
    }

	//
	// If we got here then the identified connection doesn't exist
	QLogFile::Log( "Failed to locate connection:\nID: %d", id );
}

//////////////////////////////////////////////////////////////////////
// QBindManager::UnbindAll
// Unbind all connections.
void QBindManager::UnbindAll()
{
  WIN32_GUFF();

  POSITION pos = conn_map.GetHeadPosition();
  QConnection *conn;

  while (pos != NULL)
    {
      conn = conn_map.GetNext( pos );
      conn->Unbind();
    }
}

//////////////////////////////////////////////////////////////////////
// QBindManager::Unbind
// Unbind the specified connection.
void QBindManager::Unbind( ConnID conn_id )
{
  WIN32_GUFF();

  POSITION pos = conn_map.GetHeadPosition();
  QConnection *conn;

  while ( pos != NULL )
    {
      conn = conn_map.GetNext( pos );
			if ( conn->ID() == conn_id )
				conn->Unbind();
    }
}

//////////////////////////////////////////////////////////////////////
// QBindManager::NotifySocketClose
// Notify of a close event on the given socket.
ConnID QBindManager::NotifySocketClose( SOCKET socket )
{
	WIN32_GUFF();

	//
	// Find the connection record and dump it to the log file
	POSITION pos = conn_map.GetHeadPosition();
  QConnection *conn;
  while ( pos != NULL )
    {
      conn = conn_map.GetNext( pos );
      if ( conn->Socket() == socket )
				{
					conn->Unbind();
					return conn->ID();
				}
    }	

	return UNDEF_CONN_ID;
}

//////////////////////////////////////////////////////////////////////
// QBindManager::TidyUp
// Perform housekeeping. This procedure maintains a timer which
// closes idle connections.
void QBindManager::TidyUp()
{
	//
	// Get current elapsed time
	time_t ltime;
	time( &ltime );

	//
	// Check idle time on all connections
	POSITION pos = conn_map.GetHeadPosition();
	while ( pos != NULL )
		{
			QConnection *pConn = conn_map.GetNext( pos );
			if ( pConn->Bound() && pConn->LastActive() < ( ltime - m_nIdleTimeout ) )
				{
					//
					// Connection has been idle for longer than timeout value
					pConn->Unbind();

					//
					// Log the timeout
					QLOG( "QBindManager::TidyUp");
					QLOGRAW( "Idle time out. Dropped connection:");
					pConn->Dump();
				}
		}
}

//////////////////////////////////////////////////////////////////////
// QConnMap Class
//////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

QConnMap::QConnMap()
{
}

QConnMap::~QConnMap()
{
}

//////////////////////////////////////////////////////////////////////
// QConnMap::Lookup
// Lookup the identified connection and, if found, return its
// class record.
BOOL QConnMap::Lookup( ConnID id, QConnection *&ret_conn )
{
  WIN32_GUFF();

  POSITION pos = conn_list.GetHeadPosition();
  while ( pos != NULL )
    {
      QConnection *conn = conn_list.GetNext( pos );
      if ( conn->ID() == id )
        {
          ret_conn = conn;
          return TRUE;
        }
    }

  return FALSE;    
}

//////////////////////////////////////////////////////////////////////
// QConnMap::SetAt
// Add a new connection record with the given identifier.
void QConnMap::SetAt(ConnID id, QConnection *conn)
{
  WIN32_GUFF();
  conn_list.AddTail( conn );
}

